import tempy

WHITESPACE = ''.join(sorted(' \t\r\n'))
ID_CHARS = ''.join(sorted('.:@0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz'))
OPERATOR_CHARS = ''.join(sorted('=!|&<>()^+-*/%~'))
LOCAL_LABEL_CHAR = '@'
SCOPE_COMMANDS = sorted(('scope','proc','struct'))
SCOPE_CLOSE_COMMANDS = sorted(('endscope','endproc','endstruct')) #must be 'end'+SCOPE_COMMANDS[x]
#TODO: unnamed struct must have different behavior than unnamed scope
MACRO_COMMANDS = sorted(('macro','mac'))
MACRO_CLOSE_COMMANDS = sorted(('endmacro','endmac'))
OPERATORS = sorted(('==','!=','||','&&','<=','>=','<<','>>','(',')','|','^','&','<','>','+','-','*','/','%','~','<','>','!'))

def splitScopes(text):
	elements = ['']
	lastC = ''
	for c in text:
		if c==':':
			if lastC==':':
				elements.append('')
		elif c=='.' and lastC!='': #ignore starting period, might be command
			elements.append('')
		else:
			elements[-1] += c
		lastC = c
	return elements

class TokenType:
	NONE, COMMAND, FUNCTION, ID, LABEL, NUMBER, OPERATOR, STRING, UNNAMED_LABEL, COMMENT, SPECIAL = range(11)

class Token:
	def __init__(self, type, text):
		self.type = type
		self.text = text

class TokenReader:
	def __init__(self):
		self.reset('')
	
	def readChar(self):
		if self.pos>=self.lineLen:
			return None
		c = self.line[self.pos]
		self.pos += 1
		return c
		
	def getChar(self):
		if self.pos>=self.lineLen:
			return None
		return self.line[self.pos]
		
	def reset(self,line):
		self.line = line
		self.lineLen = len(line)
		self.pos = 0
		self.startPos = 0
		
	def rewind(self,d=1):
		self.pos -= 1
			
	def readCommand(self):
		self.skipWhitespace()
		self.startPos = self.pos
		chars = ''
		fc = self.readChar()
		if fc!='.':
			raise tempy.Error('command expected')
		while True:
			c = self.readChar()
			if c==None:
				break
			if not c in ID_CHARS:
				self.rewind()
				break
			chars += c
		if len(chars)<=1:
			raise tempy.Error('command expected')
		return Token(TokenType.COMMAND, chars.lower())
		
	def readOperator(self):
		self.skipWhitespace()
		self.startPos = self.pos

		chars = self.readChar()
		if chars not in OPERATOR_CHARS:
			raise tempy.Error('operator expected')
		if self.getChar()!=None and self.getChar() in OPERATOR_CHARS:
			chars += self.readChar()

		#check if both characters together make a valid operator
		if len(chars)>1:
			if chars in OPERATORS:
				return Token(TokenType.OPERATOR, chars)
			else:
				chars = chars[0]

		#check if the first character alone makes a valid operator
		if chars == '=': #is assignment operator for constants? (special case)
			return Token(TokenType.SPECIAL, chars)
		elif chars in OPERATORS:
			return Token(TokenType.OPERATOR, chars)
		else:
			raise tempy.Error("unknown operator '"+chars+"'")

	def readId(self):
		self.skipWhitespace()
		self.startPos = self.pos
		chars = ''
		while True:
			c = self.readChar()
			if c==None:
				break
			if not c in ID_CHARS:
				self.rewind()
				break
			chars += c
		if len(chars)==0:
			raise tempy.Error('identifier expected')
		if chars[-1]==':':
			return Token(TokenType.LABEL, chars[:-1])
		else:
			return Token(TokenType.ID, chars)
	
	def readString(self):
		self.skipWhitespace()
		self.startPos = self.pos
		quoteChar = self.readChar()
		chars = ''
		escape = False
		while True:
			c = self.readChar()
			if c==None:
				return None
			if c=='\\':
				escape = not escape
			elif not escape and c==quoteChar:
				break
			chars += c
		return Token(TokenType.STRING, chars)

	def readRest(self):
		self.startPos = self.pos
		self.skipWhitespace()
		return Token(TokenType.SPECIAL, self.line[self.pos:])
	
	def skipWhitespace(self):
		while True:
			c = self.readChar()
			if c == None:
				break
			if c not in WHITESPACE:
				self.rewind()
				break
	
	def read(self):
		self.startPos = self.pos
		self.skipWhitespace()
		fc = self.getChar()
		tok = None
		if fc==None:
			pass
		elif fc=='.': #command
			tok = self.readCommand()
		elif fc==':': #unnamed label or global scope reference
			self.readChar()
			if self.getChar()==':':
				self.rewind()
				tok = self.readId() #global scope reference
		elif fc in ID_CHARS:
			tok = self.readId()
		elif fc in OPERATOR_CHARS:
			tok = self.readOperator()
		elif fc == '"' or fc=="'":
			tok = self.readString()
		elif fc == ';':
			tok = Token(TokenType.COMMENT, '')
		elif fc==',' or fc=='*':
			self.readChar()
			tok = Token(TokenType.SPECIAL, fc)
		else:
			self.readChar()
		return tok
